#! /usr/bin/env python
# 
# Merges two versions of the same file, using an earlier version as a baseline
# (i.e., 3-way merge). If the baseline version is not supplied it is determined
# as the most recent backup older than both versions.
#
# Notes:
# - The merge process can easily get confused is there are parallel changes to
#   the same sections.
# - Use a visual merge script to help with manual reconciliation (e.g., via kdiff3).
#
# TODO:
# - Make stdout the default.
# - Integrate merge heuristics for common issues.
#   -- For example, accounting for changes flagged with comments:
#      ## OLD: from regex import my_re
#      from my_regex import my_re
#   -- Accounting for changes elsewhere.
#

"""Perform 3-way merge with baseline from backup dir"""

# Standard packages
import re

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system

# Constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")
BASENAME = "basename"
BACKUP_DIR = "backup-dir"
BACKUP_FILENAME = "backup-filename"
FILENAME1 = "filename1"
FILENAME2 = "filename2"
QUIET = "quiet"
USE_STDOUT = "stdout"
SKIP_BASELINE = "skip-baseline"
UPDATE_FILE1 = "update-file1"
IGNORE_ERRORS = "ignore-errors"
MERGE = system.getenv_text("MERGE", "/usr/bin/merge -p",
                           description="Command line for merge with output piping arg (e.g., -p)")

#...............................................................................

def get_timestamp(filename):
    """Returns timestamp for FILENAME as string value"""
    # EX: get_timestamp("/vmlinuz") => "2021-04-15 04:24:54"
    result = system.get_file_modification_time(filename, as_float=False)
    debug.assertion(isinstance(result, str))
    return result


def get_numeric_timestamp(filename):
    """Returns timestamp for FILENAME as floating point value"""
    # EX: get_numeric_timestamp("/vmlinuz") => 1618478694.0
    result = system.get_file_modification_time(filename, as_float=True)
    debug.assertion(isinstance(result, float))
    return result

#...............................................................................

class Script(Main):
    """Adhoc script processing class"""
    filename1 = ""
    filename2 = ""
    backup_filename = ""
    backup_dir = "./backup"
    basename = ""
    update_file1 = False
    skip_baseline = False
    quiet = False
    use_stdout = False
    ignore_errors = False

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        self.filename1 = self.get_parsed_argument(FILENAME1)
        self.filename2 = self.get_parsed_argument(FILENAME2)
        self.backup_filename = self.get_parsed_option(BACKUP_FILENAME)
        self.backup_dir = self.get_parsed_option(BACKUP_DIR, self.backup_dir)
        self.basename = self.get_parsed_option(BASENAME, self.filename1)
        self.update_file1 = self.get_parsed_option(UPDATE_FILE1, self.update_file1)
        self.skip_baseline = self.get_parsed_option(SKIP_BASELINE, self.skip_baseline)
        self.quiet = self.get_parsed_option(QUIET)
        self.use_stdout = self.get_parsed_option(USE_STDOUT)
        self.ignore_errors = self.get_parsed_option(IGNORE_ERRORS)
        ## self.x = self.get_parsed_option(X)
        debug.trace_object(5, self, label="Script instance")

    def check_regular_file(self, filename):
        """Make sure regular file exists; otherwise, issue error and exit"""
        ok = (system.file_exists(filename) and system.is_regular_file(filename))
        if not ok and not self.ignore_errors:
            system.exit(f"Error: expecting regular file for '{filename}'")
        return ok
        
    def run_main_step(self):
        """Main processing step"""
        debug.trace_fmtd(5, "Script.run_main_step(): self={s}", s=self)

        # Validate input files
        self.check_regular_file(self.filename1)
        # TODO: if .../is_dir(self.filename2): self.filename2 = .../form_path(filename2, basename(filename2))
        self.check_regular_file(self.filename2)

        # Find the backup baseline if not specified
        if not self.backup_filename:
            EPSILON = 1e-6
            timestamp1 = get_numeric_timestamp(self.filename1)
            timestamp2 = get_numeric_timestamp(self.filename2)
            min_timestamp = min(timestamp1, timestamp2) - EPSILON

            # Find most recent backup older than two files
            SEP = "\t"
            backup_files = gh.get_files_matching_specs("{d}/{b}{SEP}{d}/{b}.*~*".
                                                       format(d=self.backup_dir, b=self.basename, SEP=SEP)
                                                       .split(SEP))
            timestamped_backups = [(f, ts) for (f, ts) in sorted(zip(backup_files,
                                                                     map(get_numeric_timestamp, backup_files)),
                                                                 key=lambda f_ts: f_ts[1], reverse=True)
                                   if (ts <= min_timestamp)]
            if timestamped_backups:
                self.backup_filename = timestamped_backups[0][0]

            # If no suitable backup, use the older of the two files
            if ((not self.backup_filename) and self.skip_baseline):
                self.backup_filename = self.filename1 if (timestamp1 < timestamp2) else self.filename2
        if not self.backup_filename:
            system.exit("Error: Unable to find baseline backup for {f1} ({ts1}) and {f2} ({ts2}) ",
                        f1=self.filename1, ts1=get_timestamp(self.filename1),
                        f2=self.filename2, ts2=get_timestamp(self.filename2))
        self.check_regular_file(self.backup_filename)

        # Do the merge and check for conflicts
        # TODO: produce report including diff listings of old vs. new
        system.create_directory(self.temp_base)
        temp_new_file = gh.form_path(self.temp_base, self.basename)
        result = gh.run("{m} {f1} {b} {f2} > {new}", m=MERGE, f1=self.filename1,
                        b=self.backup_filename, f2=self.filename2, new=temp_new_file)
        if re.search(r"conflict in merge", result):
            system.print_stderr("Error in merge:\n\t{r}", r=gh.indent_lines(result))
        elif self.update_file1:
            if not self.quiet:
                print("Updating {f1}", f1=self.filename1)
            gh.rename_file(self.filename1, (self.filename1 + ".backup"))
            gh.copy_file(temp_new_file, self.filename1)
        elif self.use_stdout:
            print(system.read_entire_file(temp_new_file))
        else:
            if not self.quiet:
                print("Merge result:\n\t{r}".format(r=result))
                debug.trace(4, f"See {temp_new_file}")
        return

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    # TODO: add examples (especially with consolidated backup directory)
    app = Script(
        description=__doc__,
        skip_input=True,
        manual_input=True,
        boolean_options=[(IGNORE_ERRORS, "Ignore errors in processing"),
                         (UPDATE_FILE1, "Replace file1 with merged result"),
                         (USE_STDOUT, "Use standard output for result"),
                         ## TODO: (X, "..."),
                         (QUIET, "Omit status messages")],
        positional_arguments=[FILENAME1, FILENAME2],
        text_options=[(BACKUP_FILENAME, "Backup file for baseline"),
                      (BACKUP_DIR, "Backup directory"),
                      (BASENAME, "Basename for backup check (if not filename1)")])
    app.run()



#! /usr/bin/env python
#
# Convenience class for regex searching, providing simple wrapper around
# static match results.
#
# Example usage:
#
#    from my_regex import my_re
#    ...
#    if (my_re.search(r"^(\d+)\:?(\d*)\s+(\d+)\:?(\d*)\s+\S.*", line)):
#        (start_hours, start_mins, end_hours, end_mins) = my_re.groups()
#
#--------------------------------------------------------------------------------
# TODO: weed out tpo_common usages (work out converter script for common cases like tpo.debug_print to debug.trace and tpo.debug_format to debug.trace_fmt)
# TODO:
# - Add examples for group(), groups(), etc.
# - Clean up script (e.g., regex => regex_wrapper).
# - Add perl-inspired accessors (e.g., PREMATCH, POSTMATCH).
#

"""Wrapper class for regex match results"""

# Standard packages
import re

# Installed packages
import six

# Local packages
from mezcla import debug
from mezcla import system
from mezcla import tpo_common as tpo

# Expose public symbols from re package, plus the wrapper class and global instance
## OLD:
## # Expose public symbols from re package,
## __all__ = re.__all__
## DEBUG: system.print_stderr("checking SKIP_RE_ALL")
RE_ALL = (not system.getenv_bool("SKIP_RE_ALL", False,
                                 "Don't use re.__all__: for sake of pylint"))
if RE_ALL:
    ## TODO: __all__ = re.__all__ + ['regex_wrapper', 'my_re']
    pass
else:
    debug.trace(4, "Omitting use of re __all__")

## TODO # HACK: make sure regex can be used as plug-in replacement 
## from from re import *

class regex_wrapper(object):
    """Wrapper class over re to implement regex search that saves match results
    note: Allows regex to be used directly in conditions"""
    # TODO: IGNORECASE = re.IGNORECASE, etc.
    # import from RE so other methods supported directly (and above constants)
    TRACE_LEVEL = debug.QUITE_DETAILED
    
    def __init__(self, ):
        tpo.debug_format("regex.__init__(): self={s}", 4, s=self)
        self.match_result = None
        # TODO: self.regex = ""

    def search(self, regex, text, flags=0, base_trace_level=None):
        """Search for REGEX in TEXT with optional FLAGS and BASE_TRACE_LEVEL (e.g., 6)"""
        ## TODO: rename as match_anywhere for clarity
        if base_trace_level is None:
            base_trace_level = self.TRACE_LEVEL
        tpo.debug_format("regex.search({r}, {t}, {f}): self={s}", (1 + base_trace_level),
                         r=regex, t=text, f=flags, s=self)
        debug.assertion(isinstance(text, six.string_types))
        self.match_result = re.search(regex, text, flags)
        if self.match_result:
            ## OLD: tpo.debug_print("match: %s" % tpo.to_string(self.match_result.groups()), 6)
            debug.trace_fmt(base_trace_level, "match: {m}; regex: {r}", m=self.grouping(), r=regex)
        return self.match_result

    def match(self, regex, text, flags=0, base_trace_level=None):
        """Match REGEX to TEXT with optional FLAGS and BASE_TRACE_LEVEL (e.g., 6)"""
        ## TODO: rename as match_start for clarity; add match_all method (wrapper around fullmatch)
        if base_trace_level is None:
            base_trace_level = 6
        tpo.debug_format("regex.match({r}, {t}, {f}): self={s}", (1 + base_trace_level),
                         r=regex, t=text, f=flags, s=self)
        self.match_result = re.match(regex, text, flags)
        if self.match_result:
            ## OLD: tpo.debug_print("match: %s" % tpo.to_string(self.match_result.groups()), 6)
            debug.trace_fmt(base_trace_level, "match: {m}; regex: {r}", m=self.grouping(), r=regex)
        return self.match_result

    def get_match(self):
        """Return match result object for last search or match"""
        result = self.match_result
        tpo.debug_format("regex.get_match() => {r}: self={s}", 7,
                         r=result, s=self)
        return result

    def group(self, num):
        """Return group NUM from match result from last search"""
        debug.assertion(self.match_result)
        result = self.match_result and self.match_result.group(num)
        tpo.debug_format("regex.group({n}) => {r}: self={s}", 7,
                         n=num, r=result, s=self)
        return result

    def groups(self):
        """Return all groups in match result from last search"""
        debug.assertion(self.match_result)
        result = self.match_result and self.match_result.groups()
        debug.trace_fmt(7, "regex.groups() => {r}: self={s}",
                        r=result, s=self)
        return result

    def grouping(self):
        """Return groups for match result or entire matching string if no groups defined"""
        # Note: this is intended to facilitate debug tracing; see example in search method above
        result = self.match_result and (self.match_result.groups() or self.match_result.group(0))
        debug.trace_fmt(7, "regex.grouping() => {r}: self={s}", r=result, s=self)
        return result

    def start(self, group=0):
        """Start index for GROUP"""
        result = self.match_result and self.match_result.start(group)
        debug.trace_fmt(7, "regex.start({g}) => {r}: self={s}", r=result, s=self, g=group)
        return result

    def end(self, group=0):
        """End index for GROUP"""
        result = self.match_result and self.match_result.end(group)
        debug.trace_fmt(7, "regex.end({g}) => {r}: self={s}", r=result, s=self, g=group)
        return result

#...............................................................................
# Initialization
#
# note: creates global instance for convenience (and backward compatibility)

my_re = regex_wrapper()

if __name__ == '__main__':
    system.print_stderr("Warning: not intended for command-line use")



#!/usr/bin/env python3
# coding: utf-8
# TODO:
# - Fix env-spec above and put coding on same line.
# - Verify the fixup for PEP-0479 didn't break anything in Python 2 (e.g., for regression testing).
#

"""Pre-processing step for text.

Example:
    >>> from mezcla.tfidf.preprocess import Preprocessor
    >>> p = Preprocessor(gramsize=2)
    >>> sorted(list([k.text for k in p.yield_keywords("a b c")]))
    ["a b", "b c"]
"""
# TPO: TODO: 
# - add logging
# - replace all_ngrams with min_ngram_size
# - add test set and check explicitly with multi-threaded access
from __future__ import absolute_import, with_statement

# Standard modules
from collections import namedtuple
## OLD: import os
import re
import sys  # python2 support
from threading import RLock

# Installed modules
from cachetools import LRUCache, cached  # python2 support
from nltk.stem import SnowballStemmer
from six.moves.html_parser import HTMLParser  # python2 support
from stop_words import get_stop_words

# Local modules
## OLD: from .dockeyword import DocKeyword
from mezcla import debug
from mezcla import system
from mezcla.tfidf.dockeyword import DocKeyword

unescape = HTMLParser().unescape

# TPO: Allow word splitting pattern to be overwritten.
# TODO: Get WORD_REGEX from environment.
# TODO: Have option to ignore words with specific punctuation (or to include).
# TODO: strip period to allow for ngrams with abbreviations (e.g., "Dr. Jones").
## OLD: import os
## OLD: import sys
## OLD:
## SPLIT_WORDS_TEXT = os.environ.get("SPLIT_WORDS", "False")
## SPLIT_WORDS = (str(SPLIT_WORDS_TEXT).upper() in ["TRUE", "1"])
## BAD_WORD_PUNCT_REGEX = os.environ.get("BAD_WORD_PUNCT_REGEX", r"[^a-zA-Z0-9_'$ -]")
SPLIT_WORDS = system.getenv_bool("SPLIT_WORDS", False,
                                 "Split by word token instead of whitespace")
BAD_WORD_PUNCT_REGEX = system.getenv_text("BAD_WORD_PUNCT_REGEX", r"[^a-zA-Z0-9_'$ -]",
                                          "Regex for punctuation not in words")
SKIP_WORD_CLEANING = system.getenv_bool("SKIP_WORD_CLEANING", False,
                                        "Omit word punctuation cleanup")

if SPLIT_WORDS:
    ## OLD: sys.stderr.write("Splitting by word token (not whitespace)\n")
    debug.trace(1, "Splitting by word token (not whitespace)\n")
    ## BAD: True
WORD_REGEX = r'\w+' if SPLIT_WORDS else r'\S+'

def handle_unicode(text):
    """Needed for the description fields."""
    if re.search(r'\\+((u([0-9]|[a-z]|[A-Z]){4}))', text):
        ## OLD: text = text.encode('utf-8').decode('unicode-escape')
        text = text.encode('utf-8', 'ignore').decode('unicode-escape', 'ignore')
    text = re.sub(r'\\n', '\n', text)
    text = re.sub(r'\\t', '\t', text)
    return text


def handle_html_unquote(text):
    """Detect if there are HTML encoded characters, then decode them."""
    if re.search(r'(&#?x?)([A-Z]|[a-z]|[0-9]){2,10};', text):
        text = unescape(text)
    return text


def handle_mac_quotes(text):
    """Handle the unfortunate non-ascii quotes OSX inserts."""
    text = text.replace(u'“', u'"').replace(u'”', u'"')\
        .replace(u'‘', u"'").replace(u'’', u"'")
    return text


def handle_text_break_dash(text):
    """Convert text break dashes into semicolons to simplify things.

    Example:
        "She loved icecream- mint chip especially"
        "She loved icecream - mint chip especially"
        both convert to
        "She loved icecream; mint chip especially"

        However,
        "The 27-year-old could eat icecream any day"
        will not be changed.
    """
    return re.sub(r'\s+-\s*|\s*-\s+', ';', text)


def clean_text(raw_text):
    """Strip text of non useful characters.
    Note: This is no-op if SKIP_WORD_CLEANING (global).
    """
    # TODO: rework so SKIP_WORD_CLEANING applied elsewhere
    text = raw_text

    
    if not SKIP_WORD_CLEANING:
        # Must strip HTML tags out first!
        text = re.sub('<[^<]+?>', ' ', text)
        text = handle_unicode(text)
        text = handle_html_unquote(text)
        text = handle_mac_quotes(text)
        text = handle_text_break_dash(text)
        text = text.lower()

        ## BAD: regex_subs = ['\t', '\n', '\r', '\s+', '&']
        regex_subs = ['\t', '\n', '\r', r'\s+', '&']
        for regex_sub in regex_subs:
            text = re.sub(regex_sub, ' ', text)
    debug.trace(7, "clean_text({raw_text}) => {text}")
    return text


def create_stemmer(language):
    """Return wordform stemmer for LANGUAGE (via SnowBall)"""
    stemmer = SnowballStemmer(language)
    
    # Define the stemmer for the particular language
    def stem_wordform(wordform):
        """Returned root of WORDFORM"""
        root = stemmer.stem(wordform)
        debug.trace_fmt(8, "stem_wordform({wf}) => {r}", wf=wordform, r=root)
        return root

    # Do sanity check
    if (language == "english"):
        debug.assertion(stem_wordform("running") == "run")

    return stem_wordform


class Preprocessor(object):
    """Prep the text for TF-IDF calculations.

    Fixes some unicode problems, handles HTML character encoding,
    and removes HTML tags.

    Strips some non alphanumeric characters, but keeps ngram boundary
    markers (eg, period (',') and semi-colon (';'))

    If a stopwords file is provided, it will remove stopwords.

    Example:
        >>> processor = Preprocessor('english_stopwords.txt')
        >>> processor.clean('He was an interesting fellow.')
        "was interesting fellow."
    """

    stopwords = set()
    contractions = r"(n't|'s|'re|'ll)$"
    negative_gram_breaks = r'[^:;!^,\?\.\[|\]\(|\)"`]+'
    supported_languages = (
        'danish', 'dutch', 'english', 'finnish', 'french', 'german', 'hungarian',
        'italian', 'kazakh', 'norwegian', 'porter', 'portuguese', 'romanian',
        'russian', 'spanish', 'swedish', 'turkish'
    )

    def __init__(self, language=None, gramsize=1, all_ngrams=None, min_ngram_size=None,
                 stopwords_file=None, stemmer=None):
        """Preprocessor must be initalized for use if using stopwords.

        stopwords_file (filename): contains stopwords, one per line
        stemmer (function):  takes in a word and returns the stemmed version
        gramsize (int=1): maximum word size for ngrams
        all_ngrams (bool):
            if true, all possible ngrams of length "gramsize" and smaller will
            be examined. If false, only ngrams of _exactly_ length "gramsize"
            will be run.
            Note: deprecated (use min_ngram_size instead).
        min_ngram_size (int-or-None):
            if integral, gives the lower bound unless all_ngrams specified.
        negative_gram_breaks (regex):
            if a word ends with one of these characters, an
            ngram may not cross that. Expressed as a _negative_ regex.
            Example:
            in the sentence "Although he saw the car, he ran across the street"
            "car he" may not be a bi-gram
        stopwords_file (filename):
            Provide a list of stopwords. If used in addition to "language", the
            provided stopwords file overrides the default.
        stemmer (function):
            A function that takes in a single argument (str) and returns a string
            as the stemmed word. Overrides the default behavior if specified.
            Default None:
                Use the NLTK snowball stemmer for the sepcified language. If
                language is not found, no stemming will take place.
        """
        self.__stemmer = None
        if language:
            ## OLD: assert language in self.supported_languages
            debug.assertion(language in self.supported_languages)
            if language in SnowballStemmer.languages:
                ## OLD: sb_stemmer = SnowballStemmer(language)
                ## self.__stemmer = sb_stemmer.stem
                self.__stemmer = create_stemmer(language)
            ## OLD:
            ## else:
            ##     debug.trace(3, "Warning: defining no-op stemmer in Preprocessor")
            ##     self.__stemmer = lambda x: x  # no change to word
            self.stopwords = get_stop_words(language)
        if stopwords_file:
            self._load_stopwords(stopwords_file)
        if stemmer:
            self.__stemmer = stemmer
        if not self.__stemmer:
            debug.trace(3, "Warning: defining no-op stemmer in Preprocessor")
            self.__stemmer = lambda x: x  # no change to word
        self.__gramsize = gramsize
        self.__all_ngrams = all_ngrams
        self.__min_ngram_size = (min_ngram_size or gramsize)

    @property
    def gramsize(self):
        """Number of words in the ngram."""
        return self.__gramsize

    @property
    def all_ngrams(self):
        """True if ngrams of size "gramsize" or smaller will be generated.

        False if only ngrams of _exactly_ size "gramsize" are generated.
        """
        return self.__all_ngrams

    @property
    def min_ngram_size(self):
        """If non-Null, can be used to override all_ngrams (q.v.), so
        ngrams go from [M, N] rather than [1, N]
        """
        return self.__min_ngram_size

    def _load_stopwords(self, filename):
        with open(filename) as f:
            words = []
            for line in f:
                words.append(line.strip())
        self.stopwords = set(words)

    def handle_stopwords(self, text):
        """Remove stop words from the text."""
        out = []
        for word in text.split(' '):
            # Remove common contractions for stopwords when checking list
            check_me = re.sub(self.contractions, '', word)
            if check_me in self.stopwords:
                continue
            out.append(word)
        return ' '.join(out)

    def normalize_term(self, text):
        """Cleans the text characters (optional), removes stopwords, and applies stemming

        Assumes the input is already the number of words you want for the ngram.
        Note: Omits word punctuation cleanup if SKIP_WORD_CLEANING (global).
        """
        text = clean_text(text)
        text = self.handle_stopwords(text)
        return self.stem_term(text)

    ## OLD: @cached(LRUCache(maxsize=10000))
    ## TODO: do performance tests to select suitable cache size
    @cached(LRUCache(maxsize=100000), lock=RLock())
    def _stem(self, word):
        ## OLD: """The stem cache is used to cache up to 10,000 stemmed words.
        """The stem cache is used to cache up to 100,000 stemmed words.

        This substantially speeds up the word stemming on larger documents.
        """
        return self.__stemmer(word)

    def stem_term(self, term):
        """Apply the standard word procesing (eg stemming). Returns a stemmed ngram."""
        return ' '.join([self._stem(x) for x in term.split(' ')])

    def yield_keywords(self, raw_text, document=None):
        """Yield keyword objects as mono, di, tri... *-grams.

        Use this as an iterator.

        Will not create ngrams across logical sentence breaks.
        Example:
            s = "Although he saw the car, he ran across the street"
            the valid bigrams for the sentences are:
            ['Although he', 'saw the', 'he saw', 'the car',
            'he ran', 'across the', 'ran across', 'the street']
            "car he" is not a valid bi-gram

        This will also stem words when applicable.
        Example:
            s = "All the cars were honking their horns."
            ['all', 'the', 'car', 'were', 'honk', 'their', 'horn']
        """
        if sys.version_info[0] < 3:  # python2 support
            if isinstance(raw_text, str):
                raw_text = raw_text.decode('utf-8', 'ignore')
        ## TPO: HACK: support min ngram size
        ## OLD: min_ngram_size = self.min_ngram_size if not self.all_ngrams else 1
        ## OLD: gramlist = range(1, self.gramsize + 1) if self.all_ngrams else [self.gramsize]
        gramlist = None
        if self.all_ngrams:
            gramlist = range(1, self.gramsize + 1)
        if self.min_ngram_size:
            debug.assertion(not gramlist)
            if gramlist:
                sys.stderr.write("Error: ignoring obsolete all_ngrams and using min_ngram_size\n")
            gramlist = range(self.min_ngram_size, self.gramsize + 1)
        if not gramlist:
            gramlist = [self.gramsize]
        ## DEBUG: sys.stderr.write("gramlist={gl}\n".format(gl=gramlist))
        debug.trace_fmt(7, "gramlist={gl}\n", gl=gramlist)

        for sentence in positional_splitter(self.negative_gram_breaks, raw_text):
            ## OLD: words = [x for x in positional_splitter(r'\S+', sentence.text)]
            ## OLD: words = [x for x in positional_splitter(WORD_REGEX, sentence.text)]
            words = positional_splitter(WORD_REGEX, sentence.text)
            # Remove all stopwords
            words_no_stopwords = []
            for w in words:
                # Remove common contractions for stopwords when checking list
                check_me = re.sub(self.contractions, '', w.text)
                if check_me not in self.stopwords:
                    words_no_stopwords.append(w)

            # Make the ngrams
            # TODO: make sure stripped stopwords block ngram (e.g. "dog and cat" =/=> "dog cat")
            for gramsize in gramlist:
                # You need to try as many offsets as chunk size
                # TODO: Allow stopwords in middle of ngram of size 3 or more
                for offset in range(0, gramsize):  # number of words offest
                    data = words_no_stopwords[offset:]
                    text_in_chunks = [data[pos:pos + gramsize]
                                      for pos in range(0, len(data), gramsize)
                                      if len(data[pos:pos + gramsize]) == gramsize]
                    for word_list in text_in_chunks:
                        word_text = ' '.join([self.stem_term(w.text) for w in word_list])
                        if (SPLIT_WORDS or (not re.search(BAD_WORD_PUNCT_REGEX, word_text))):
                            word_global_start = sentence.start + word_list[0].start
                            word_global_end = sentence.start + word_list[-1].end
                            ## OLD: yield DocKeyword(word_text, document=document,
                            ##                       start=word_global_start, end=word_global_end)
                            try:
                                yield DocKeyword(word_text, document=document, start=word_global_start, end=word_global_end)
                            except StopIteration:
                                pass
        ## Ye olde introduce-a-stupid-feature at M.m (rather than M.0)--gotta hate Python!
        ##    https://www.python.org/dev/peps/pep-0479.
        ## OLD: raise StopIteration
        return


PositionalWord = namedtuple('PositionalWord', ['text', 'start', 'end'])


def positional_splitter(regex, text):
    r"""Yield sentence chunks (as defined by the regex) as well as their location.

    NOTE: the regex needs to be an "inverse match"
    Example:
        To split on whitespace, you match:
        r'\S+'  <-- "a chain of anything that's NOT whitespace"
    """
    for res in re.finditer(regex, text):
        ## OLD: yield PositionalWord(res.group(0), res.start(), res.end())
        try:
            yield PositionalWord(res.group(0), res.start(), res.end())
        except StopIteration:
            pass
    ## Same as above
    ## OLD: raise StopIteration
    return


def main():
    """Entry point for script: just runs a simple test"""
    text = "my man fran is not a man"
    p = Preprocessor(gramsize=2)
    debug.trace_object(4, p)
    ngrams = list([p.text for p in p.yield_keywords(text)])
    debug.assertion("my man" in ngrams)
    debug.assertion("my man fran" not in ngrams)
    debug.assertion("fran man" not in ngrams)
    debug.trace_expr(4, ngrams)
    
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone. A simple test willl be run.")
    main()



#! /usr/bin/env python
#

"""Simple assertion that doesn't abort upon failure"""

import inspect
from mezcla import debug
from mezcla import system

if __debug__:

    def soft_assertion(expression):
        """Issue warning if EXPRESSION is not True"""
        debug.trace_expr(5, expression)
        if not expression:
            debug.trace(4, "expression failed")
            caller = inspect.stack()[1]
            (_frame, filename, line_number, _function, _context, _index) = caller
            system.print_error(f"Warning: assertion failed at {filename}:{line_number}\n")

else:

    def soft_assertion(_expression):
        """no-op for assertion check"""
        pass                             # pylint: disable=unnecessary-pass


def main():
    """Entry point for script"""
    system.print_error("Warning: not intended for command line usage.")
    debug.trace(1, "Examples follow.")
    soft_assertion(0 <= debug.get_level() <= 10)
    soft_assertion((2 + 2) == 5)

if __name__ == '__main__':
    main()



#! /usr/bin/env python
#
# Utility functions for doing data analysis, such as wrappers for functions
# provided by pandas and sklearn.
#
# Note:
# - This started out as way to provide commonly used options for CSV reading
#   with Pandas, along with tracing.
# - It is a little idiosyncratic with a bias towards Unix assumptions
#   (e.g., comments more likely in CSV files).
#
# TODO:
# - Add simple wrapper class for commonly used idioms, sich as len(df.columns) for number of columns.
#

"""Utility functions for work with data (e.g., pandas wrappers)"""

# Standard module
import pandas as pd

# Local modules
## OLD:
## import system
## import debug
from mezcla import debug
from mezcla import system

# Constants
# Note: Delim defaults to None so that dialect inference can be used.
# This is a quirk with Pandas compared to the cvs module.
## OLD: DELIM = system.getenv_text("DELIM", ",",
DELIM = system.getenv_value("DELIM", None,
                            "Delimiter for input and output tables")
COMMENT = 'comment'
DIALECT = 'dialect'
EXCEL = 'excel'

#-------------------------------------------------------------------------------

def read_csv(filename, **in_kw):
    """Wrapper around pandas read_csv"""
    ## TODO: clarify dtype usage
    ## BAD
    ## kw = {'comment': "#", 'delimiter': DELIM, 'dtype': str,
    ##       'error_bad_lines': False, 'escapechar': '\\', 'keep_default_na': False}
    ## OLD:
    ## kw = {'comment': "#", 'delimiter': DELIM, 'dtype': str,
    ##       'error_bad_lines': False, 'keep_default_na': False}
    kw = {'delimiter': DELIM, 'dtype': str,
          'error_bad_lines': False, 'keep_default_na': False}
    kw.update(**in_kw)
    # Add special processing (n.b., a bit idiosyncratic)
    if ((COMMENT not in kw) and (kw.get(DIALECT) != EXCEL)):
        debug.trace_fmt(4, "Enabling comments in read_csv")
        kw[COMMENT] = "#"
    ## OLD:
    ## debug.trace_fmt(5, "read_csv({f}, cmt={comment}, del={delimiter}, dty={dtype},"
    ##                 "ebl={error_bad_lines}, kdn={keep_default_na}, in_kw={ikw})", f=filename, **kw, ikw=in_kw)
    debug.trace_fmt(5, "read_csv({f}, [in_kw={ikw}])", f=filename, ikw=in_kw)
    debug.trace_fmt(6, "\tkw={k}", k=kw)
    df = pd.read_csv(filename, **kw)
    debug.trace(4, f"read_csv({filename}) => {df}")
    return df


def to_csv(filename, data_frame):
    """Output to FILENAME the CSV for DATA_FRAME without index column"""
    result = None
    try:
        result = data_frame.to_csv(filename, index=False)
    except:
        debug.trace(4, f"Exception during write_csv: {system.get_exception()}")
    debug.trace(4, f"to_csv({filename}, {data_frame}) => {result}")
    return result
#
write_csv = to_csv

def lookup_df_value(data_frame, return_field, lookup_field, lookup_value):
    """Return value for DATA_FRAME's RETURN_FIELD given LOOKUP_FIELD value LOOKUP_VALUE"""
    ## TODO: rework in terms of Pandas primitives
    value = None
    try:
        # TODO: trace out index location
        matches = [row[return_field] for index, row in data_frame.iterrows() 
                   if (row[lookup_field] == lookup_value)]
        if matches:
            value = matches[0]
    except:
        debug.trace(4, f"Exception during lookup_df_value: {system.get_exception()}")
    debug.trace(7, f"lookup_df_value(_, {return_field}, {lookup_field}, {lookup_value}) => {value}")
    return value


def main():
    """Entry point for script"""
    ## OLD: debug.trace(2, "Warning: Not intended to being invoked directly")
    system.print_stderr("Error: Not intended to being invoked directly")
    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    main()
